#version 150

#include "/include.glsl"

uniform sampler2D gtexture;
uniform sampler2D lightmap;
uniform sampler2D shadowtex0;
uniform sampler2D shadowtex1;
uniform sampler2D shadowcolor0;
uniform sampler2D normals;
uniform sampler2D specular;

/* RENDERTARGETS:0,2,3,4,5 */
out vec4 outColor0;
out vec4 outColor1;
out vec4 outColor2;
out vec4 outColor3;

in float depth;
in vec2 texCoord;
in vec4 shadowUv;
in vec2 screenUv;
in float bias;
in vec3 foliageColor;
in vec2 lightMapCoords;
in vec4 tangent;
in vec3 geoNormal;
in vec3 viewDirection;
in vec3 lightDirection;

#define labPBR

void main() {
    // Get fragment color and transperency
    vec4 outputColorData = texture(gtexture, texCoord);

    // Should be self explanitory, set transperency to the transperency value
    float transperency = outputColorData.a;

    // Ignore clear fragments
    #if View != 1
        if (transperency < alphaTestRef) {
            discard;
        }
    #endif
    
    // FOG
    float fogPercent = 0.0;
    float depthInBlocks = depth * far;
    float fogDistanceMultiplier = 1.0;

    if (isEyeInWater == 1) {
        fogDistanceMultiplier = 0.1;
    } else if (isEyeInWater == 2 || isEyeInWater == 3) {
        fogDistanceMultiplier = 0.01;
    }

    bool clipFogNew;

    #ifdef ClipFog
        if (isEyeInWater != 0) {
            clipFogNew = false;
        } else {
            clipFogNew = true;
        }
    #else
        clipFogNew = false;
    #endif

    #if FogType == 1
        // Smoothstep
        fogPercent = smoothstep(SmoothstepFogStartPercent * fogDistanceMultiplier / 100.0, SmoothstepFogEndPercent * fogDistanceMultiplier / 100.0, depth);
    #elif FogType == 2
        // Linear
        float fogStartBlocks = (LinearFogStartPercent / 100.0) * far;
        float fogEndBlocks = (LinearFogEndPercent / 100.0) * far;

        depthInBlocks -= fogStartBlocks;
        depthInBlocks /= fogEndBlocks;

        fogPercent = clamp(depthInBlocks, 0, 1);
    #elif FogType == 3
        // Exponential
        fogPercent = 1.0 - clamp(exp(-(ExpDensity / 100.0) * depth), 0, 1);
    #elif FogType == 4
        // Exponential^2
        const float LOG2 = -1.442695;
        float d = (Exp2Density / 100.0) * depth;
        fogPercent = 1.0 - clamp(exp2(d * d * LOG2), 0.0, 1.0);
    #endif

    if (clipFogNew && fogPercent == 1.0) {
        discard;
    }

    #ifdef labPBR
        // Normals
        vec4 normalData = texture(normals, texCoord) * 2.0 - 1.0;

        vec3 shadowLightDirection = normalize(mat3(gbufferModelViewInverse) * shadowLightPosition);

        vec3 normal = vec3(normalData.xy, sqrt(1.0 - dot(normalData.xy, normalData.xy)));
        mat3 TBN = tbnNormalTangent(mat3(gbufferModelViewInverse) * geoNormal, mat3(gbufferModelViewInverse) * tangent.xyz);
        vec3 normalWorldSpace = TBN * normal;

        vec4 biasOffset = shadowProjection * vec4(mat3(shadowModelView) * (mat3(gbufferModelView) * (normalWorldSpace)), 1);
        vec4 newShadowUv = vec4(shadowUv.xyz - (biasOffset.xyz / biasOffset.w * bias), shadowUv.w);

        float lightBrightness = clamp(dot(shadowLightDirection, normalWorldSpace), 0, 1);

        vec4 normalShadowUv = vec4(shadowUv.xyz, lightBrightness);

        // Specular
        vec4 specularData = texture(specular, texCoord);

        float roughness = pow(1.0 - specularData.r, 2.0);
        float f0 = specularData.g;
        vec3 reflectedDirection = reflect(normalize(-lightDirection), normalize((normalWorldSpace)));

        float specularAmount = pow(max(dot(viewDirection, reflectedDirection), 0.0), 100.0);
        vec3 specularColor = (1.0 - roughness) * specularAmount * vec3(1);

        if (f0 == 0) {
            specularColor = vec3(0);
        }
    #else
        vec4 normalShadowUv = shadowUv;
        vec3 normalWorldSpace = geoNormal;
        vec3 specularColor = vec3(0);
        float roughness = 1.0;
    #endif

    vec2 lmcoord = lightMapCoords;

    bool hasColor = false;

    outputColorData += vec4(entityColor.rgb, 0);

    // Get light color based off lightmap coords
    vec3 lightColor = texture(lightmap, lmcoord).rgb;

    float inSkylight = 0.0;
    float sunInSky = clamp(dot(upPosition, normalize(sunPosition)), 0, 1);

    #ifdef ShadowsEnabled
        float closestDepth = texture(shadowtex0, normalShadowUv.xy).x;

        float usedDepth = closestDepth;

        #ifdef ColorShadowsEnabled
            float closestOpaqueDepth = texture(shadowtex1, normalShadowUv.xy).x;
            usedDepth = closestOpaqueDepth;
        #endif
        
        float currentDepth = normalShadowUv.z;

        float actualShadowBrightness = 0.05 + (ShadowBrightness * (1.0 - 0.05));

        if (usedDepth < currentDepth) {
            // In shadow
            lmcoord.y = min(lmcoord.y, ((ShadowBrightness * 30.0) + 1.0) / 32.0);
            lightColor = texture(lightmap, lmcoord).rgb;
            specularColor = vec3(0);
        } else {
            // Not in shadow
            inSkylight = 0.5;
            lmcoord.y = min(lmcoord.y, mix(((ShadowBrightness * 30.0) + 1.0) / 32.0, 31.0 / 32.0, normalShadowUv.w));
            #ifdef ColorShadowsEnabled
                if (closestDepth < currentDepth) {
                    vec4 shadowLightColor = texture2D(shadowcolor0, normalShadowUv.xy);
                    #ifndef DarknessPitchBlack
                        shadowLightColor *= ShadowBrightness;
                    #endif
                    shadowLightColor.rgb = mix(vec3(1.0), shadowLightColor.rgb, shadowLightColor.a);
                    shadowLightColor.rgb = mix(shadowLightColor.rgb, vec3(1), clamp(((((lmcoord.x - (1.0 / 16.0)) * (255.0 / 231.0)) + 0.5) * -(sunInSky - 1.0)) + ((((lmcoord.x - (1.0 / 16.0)) * (255.0 / 231.0)) + 0.15) * sunInSky), 0, 1));
                    #ifdef DarknessPitchBlack
                        #ifdef BlackShadowsAtNight
                            shadowLightColor *= ShadowBrightness;
                        #endif
                    #endif
                    lightColor = shadowLightColor.rgb * texture(lightmap, lmcoord).rgb;
                    hasColor = true;
                } else {
                    lightColor = texture(lightmap, lmcoord).rgb;
                }
            #else
                lightColor = texture(lightmap, lmcoord).rgb;
            #endif
        }
    #endif

    // Set light values
    float blockLight = (lmcoord.x - (1.0 / 16.0)) * (255.0 / 231.0);
    float skyLight = (lmcoord.y - (1.0 / 16.0)) * (255.0 / 231.0);

    // min light == 0.03125    max light == 0.96875

    #ifdef DarknessPitchBlack
        lmcoord.x *= step(0.001, max(blockLight, (1.0 / 16.0)));
        lmcoord.y *= step(0.001, max(skyLight, (1.0 / 16.0)));

        float isLit = clamp(blockLight + skyLight, min(blockLight, skyLight), 1.0);

        float nightDarknessSmoothness = 10.0;

        #ifdef BlackShadowsAtNight
            vec3 newLight = texture(lightmap, lmcoord).rgb * isLit * clamp(sunInSky + blockLight + inSkylight, 0, 1);
        #else
            vec3 newLight = texture(lightmap, lmcoord).rgb * isLit * clamp(sunInSky + blockLight + skyLight, 0, 1);
        #endif

        lightColor = mix(newLight, lightColor, hasColor);
    #endif
    
    #ifdef EnableGreyscale
        #ifdef UseGreyscaleInsteadOfLighting
            vec3 outputColor = outputColorData.rgb * foliageColor;
        #else
            vec3 outputColor = outputColorData.rgb * foliageColor * (lightColor + specularColor);
        #endif
    #else
        vec3 outputColor = outputColorData.rgb * foliageColor * (lightColor + specularColor);
    #endif

    vec3 usedFogColor = fogColor;

    #ifndef SkyAlwaysGreyscale
        usedFogColor = vec3(getLuminosity(usedFogColor));
    #endif

    // Apply fog
    outputColor = mix(outputColor, usedFogColor, fogPercent);

    // Output fragment color
    outColor0 = vec4(outputColor, transperency);

    // Output fragment data
    outColor1 = vec4(blockLight, skyLight, abs(normalWorldSpace.xy));
    outColor2 = vec4(normalWorldSpace.xyz, abs(normalWorldSpace.z));
    outColor3 = vec4(texture(lightmap, lmcoord).rgb, 0);
}